      Program SyncChkM
!
!     Check synchronization of system times between systems connected
!     by shared memory.
!     MASTER process.
!
!     Written by:  Mark Paulk
!     System Development Corporation
!     22 September 1983
!

      include 'syncchk.cmn'

      character*4 this_node, other_node, shmem
      character*1 answer
      integer input_addr(2), output_addr(2)
      logical sma /.false./, smb /.false./


!     Log onto the network to submit the other jobs to synch.

      status = d$signon ('$SYS')
      if (.not. status) call lib$signal(%val(status))

!     Find out where the program is running.

      status = sys$trnlog ('$PROCESSOR', , this_node, , , %val(6))
      if (.not. status) call lib$signal(%val(status))

      if (this_node .eq. 'VX01') then
         sma = .true.
      else if (this_node .eq. 'VX02') then
         sma = .true.
      else if (this_node .eq. 'VX03') then
         sma = .true.
         smb = .true.
      else if (this_node .eq. 'VX04') then
         sma = .true.
         smb = .true.
      else if (this_node .eq. 'VX05') then
         sma = .true.
         smb = .true.
      else if (this_node .eq. 'VX06') then
         sma = .true.
         smb = .true.
      else
         type *, this_node, ' is not connected to shared memory.'
         stop ' '
      end if

!     Find out which node is to be checked.

      write (6,50)
 50   format ('$Node to check? ')
      read (5,60) j, other_node
 60   format (q,a)
      call str$upcase (other_node, other_node)
      if (other_node .eq. 'VX01' .and. sma) then
         shmem = '$SMA'
      else if (other_node .eq. 'VX02' .and. sma) then
         shmem = '$SMA'
      else if (other_node .eq. 'VX03' .and. sma) then
         shmem = '$SMA'
         smb = .false.
      else if (other_node .eq. 'VX03' .and. smb) then
         shmem = '$SMB'
         sma = .false.
      else if (other_node .eq. 'VX04' .and. sma) then
         shmem = '$SMA'
         smb = .false.
      else if (other_node .eq. 'VX04' .and. smb) then
         shmem = '$SMB'
         sma = .false.
      else if (other_node .eq. 'VX05' .and. smb) then
         shmem = '$SMB'
      else if (other_node .eq. 'VX06' .and. smb) then
         shmem = '$SMB'
      else
         type *, this_node, ' and ', other_node,
     &   ' are not interconnected via shared memory.'
         stop ' '
      end if

!     Define flag characteristics for the global section as Global,
!     Permanent, and Writeable.

      flags = SEC$M_GBL .or. SEC$M_PERM .or. SEC$M_WRT

!     Open the disk file for global section creation and mapping by calling
!     the bliss routine OPENSEC.
!     Define the input addresses.
!     Create and map a global section in shared memory.
!     Initialize the data in the shared memory.

      call opensec ('CHECK.SYN', sec_size, channel)
      input_addr(1) = %loc(buffer(1))
      input_addr(2) = %loc(buffer(byte_size))
      status = sys$crmpsc (input_addr, output_addr, , %val(flags),
     &shmem//':SYNCHECK', , , %val(channel), %val(sec_size), , , )
      write (6,70) shmem, input_addr, output_addr
 70   format (1x,a,' - Specified:', 2z8,
     &' Obtained:', 2z8)
      if(.not. status) call lib$stop(%val(status))

!     Initialize.

      do i=1,n_checks
         time(i) = 0
      end do
      do i=1,2
         local_time_start(i) = 0
         local_time_end(i) = 0
         remote_time_start(i) = 0
         remote_time_end(i) = 0
      end do

!     Start the slave process.

      status = d$submit (other_node, 'SYSTEM',
     &'ARC$SYSTEM:SYNCCHKS', shmem)
      if (.not. status) call lib$signal(%val(status))

!     Call a kernel mode routine to check the clocks.

      call sys$gettim (local_time_start)
      status = sync_check_master (time, n_checks/2)
      if (.not. status) call lib$signal(%val(status))
      call sys$gettim (local_time_end)

!     Print the results.

      write (6,80)
      write (99,80)
 80   format (' Remote clock      Local clock     Difference')
d      total = 0
      do i=1,n_checks-1,2
         diff = time(i)-time(i+1)
d         total = total + diff
         write (6,90) time(i), time(i+1), diff
         write (99,90) time(i), time(i+1), diff
 90      format (1x, i12, 5x, i11, 5x, i8)
      end do
d      avg = total / (n_checks/2)
d      type *, 'Average = ', avg
      write (6,100) local_time_start(2), local_time_start(1),
     &local_time_end(2), local_time_end(1),
     &remote_time_start(2), remote_time_start(1),
     &remote_time_end(2), remote_time_end(1)
      write (99,100) local_time_start(2), local_time_start(1),
     &local_time_end(2), local_time_end(1),
     &remote_time_start(2), remote_time_start(1),
     &remote_time_end(2), remote_time_end(1)
100   format (' Local start:  ', 2z8.8, '  Stop: ', 2z8.8/
     &' Remote start: ', 2z8.8, '  Stop: ', 2z8.8)

!     Cleanup.

      status = sys$dassgn(%val(channel))
      status = sys$dgblsc(,shmem//':SYNCHECK',)

      stop
      end
